import torch
import torch.multiprocessing as mp
from typing import Optional, Tuple

from base import BaseSampler, NegativeSampling
from neighbor_sampler import NeighborSampler

class RandomWalkSampler(BaseSampler):
    def __init__(
        self,
        edge_index: torch.Tensor,
        num_nodes: int,
        num_layers: int,
        workers = 1
    ) -> None:
        r"""__init__
        Args:
            edge_index: all edges in the graph
            num_nodes: the num of all nodes in the graph
            num_layers: the num of layers to be sampled
            fanout: the list of max neighbors' number chosen for each layer
            workers: the number of threads, default value is 1
        """
        super().__init__()
        self.sampler = NeighborSampler(edge_index, num_nodes, num_layers, 
                                       [1 for _ in range(num_layers)], workers)
        self.num_layers = num_layers
        # 线程数不超过torch默认的omp线程数
        self.workers = min(workers, torch.get_num_threads())
    
    def sample_from_nodes(
        self,
        nodes: torch.Tensor
    ) -> Tuple[torch.Tensor, list]:
        r"""Performs mutilayer sampling from the nodes specified in: nodes
        The specific number of layers is determined by parameter: num_layers
        returning a sampled subgraph in the specified output format: Tuple[torch.Tensor, list].

        Args:
            nodes: the list of seed nodes index
        Returns:
            sampled_nodes: the node sampled
            sampled_edge_index: the edge sampled
        """
        return self.sampler.sample_from_nodes(nodes)
    
    def sample_from_edges(
        self,
        edges: torch.Tensor,
        edge_label: Optional[torch.Tensor] = None,
        neg_sampling: Optional[NegativeSampling] = None
    ) -> Tuple[torch.Tensor, list]:
        r"""Performs sampling from the edges specified in :obj:`index`,
        returning a sampled subgraph in the specified output format.

        Args:
            edges: the list of seed edges index
            edge_label: the label for the seed edges.
            neg_sampling: The negative sampling configuration
        Returns:
            sampled_nodes: the nodes sampled
            sampled_edge_index_list: the edges sampled
        """
        return self.sampler.sample_from_edges(edges, edge_label, neg_sampling)
    
if __name__=="__main__":
    edge_index1 = torch.tensor([[0, 1, 1, 1, 2, 2, 2, 4, 4, 4, 5], # , 3, 3
                                [1, 0, 2, 4, 1, 3, 0, 3, 5, 0, 2]])# , 2, 5
    num_nodes1 = 6
    # Run the random walk sampling
    sampler=RandomWalkSampler(edge_index=edge_index1, num_nodes=num_nodes1, num_layers=3, workers=4)

    sampled_nodes, sampled_edge_index = sampler.sample_from_nodes(torch.tensor([1,2]))
    
    # Print the result
    print('sampled_nodes_id: \n',sampled_nodes, '\nedge_index: \n',sampled_edge_index)
