#include <iostream>
#include <set>
#include <omp.h>
#include <pybind11/pybind11.h>
#include <pybind11/numpy.h>
#include <pybind11/stl.h>

using namespace std;
namespace py = pybind11;

typedef int NodeIDType;
// typedef int EdgeIDType;
// typedef float TimeStampType;

class TemporalNeighborBlock;
class TemporalGraphBlock;
TemporalNeighborBlock get_neighbors(vector<NodeIDType>& row, vector<NodeIDType>& col, int num_nodes, int threads);
TemporalGraphBlock neighbor_sample_from_node(NodeIDType node, vector<NodeIDType>& neighbors, int deg, int fanout, int threads);
TemporalGraphBlock neighbor_sample_from_nodes(vector<NodeIDType>& nodes, vector<vector<NodeIDType>>& neighbors, vector<NodeIDType>& deg, int fanout, int threads);
vector<NodeIDType> heads_unique(vector<NodeIDType>& array, vector<NodeIDType>& heads);

template<typename T>
inline py::array vec2npy(const std::vector<T> &vec)
{
    // need to let python garbage collector handle C++ vector memory 
    // see https://github.com/pybind/pybind11/issues/1042
    // non-copy value transfer
    auto v = new std::vector<T>(vec);
    auto capsule = py::capsule(v, [](void *v)
                               { delete reinterpret_cast<std::vector<T> *>(v); });
    return py::array(v->size(), v->data(), capsule);
    // return py::array(vec.size(), vec.data());
}

/* 
 * NeighborSampler Utils
 */
class TemporalNeighborBlock
{
    public:
        std::vector<vector<NodeIDType>*> neighbors;
        std::vector<int> deg;

        TemporalNeighborBlock(){}
        TemporalNeighborBlock(std::vector<vector<NodeIDType>*>& neighbors, 
                              std::vector<int> &deg):
                              neighbors(neighbors), deg(deg){}
        py::array get_node_neighbor(int node_id){
            return vec2npy(*(neighbors[node_id]));
        }
        int get_node_deg(int node_id){
            return deg[node_id];
        }
};

class TemporalGraphBlock
{
    public:
        std::vector<NodeIDType> row;
        std::vector<NodeIDType> col;
        std::vector<NodeIDType> nodes;

        TemporalGraphBlock(){}
        TemporalGraphBlock(std::vector<NodeIDType> &_row, std::vector<NodeIDType> &_col,
                           std::vector<NodeIDType> &_nodes):
                           row(_row), col(_col), nodes(_nodes){}
};

TemporalNeighborBlock get_neighbors(
        vector<NodeIDType>& row, vector<NodeIDType>& col, int num_nodes, int threads){
    int edge_num = row.size();
    TemporalNeighborBlock tnb = TemporalNeighborBlock();
    tnb.deg.resize(num_nodes, 0);
    double start_time = omp_get_wtime();
#pragma omp parallel for num_threads(threads)
    for(int i=0; i<num_nodes; i++)
        tnb.neighbors.push_back(new vector<NodeIDType>());
#pragma omp parallel for num_threads(threads)
    for(int i=0; i<edge_num; i++){
        //ڵھ
        tnb.neighbors[row[i]]->push_back(col[i]);
        //ڵ
        tnb.deg[row[i]]++;
    }
    double end_time = omp_get_wtime();
    cout<<"get_neighbors consume: "<<end_time-start_time<<"s"<<endl;
    return tnb;
}

TemporalGraphBlock neighbor_sample_from_nodes(
        vector<NodeIDType>& nodes, vector<vector<NodeIDType>>& neighbors, 
        vector<NodeIDType>& deg, int fanout, int threads){    
    TemporalGraphBlock tgb = TemporalGraphBlock();
    double start_time = omp_get_wtime();
#pragma omp parallel for num_threads(threads)
    for(int i=0; i<nodes.size(); i++){
        NodeIDType node = nodes[i];
        TemporalGraphBlock tgb_i = neighbor_sample_from_node(node, neighbors[node], deg[node], fanout, threads);
        tgb.row.insert(tgb.row.end(),tgb_i.row.begin(),tgb_i.row.end());
        tgb.col.insert(tgb.col.end(),tgb_i.col.begin(),tgb_i.col.end());
    }
    double end_time = omp_get_wtime();
    cout<<"neighbor_sample_from_nodes consume: "<<end_time-start_time<<"s"<<endl;
    //sampled nodes ȥ
    start_time = end_time;
    tgb.nodes.assign(tgb.col.begin(), tgb.col.end());
    heads_unique(tgb.nodes, nodes);
    // cout<<"nodes: "<<tgb.nodes.size()<<endl;
    end_time = omp_get_wtime();
    cout<<"unique consume: "<<end_time-start_time<<"s"<<endl;

    return tgb;
}

/*-------------------------------------------------------------------------------------**
**------------Utils--------------------------------------------------------------------**
**-------------------------------------------------------------------------------------*/

TemporalGraphBlock neighbor_sample_from_node(
        NodeIDType node, vector<NodeIDType>& neighbors, 
        int deg, int fanout, int threads){
    TemporalGraphBlock tgb = TemporalGraphBlock();
    srand((int)time(0));
    if(deg>fanout){
        //ȴȳĻҪѡfanoutھ
#pragma omp parallel for num_threads(threads)
        for(int i=0; i<fanout; i++){
            //ѭѡfanoutھ
            auto chosen_iter = neighbors.begin() + rand()%(deg-i);
            tgb.col.push_back(*chosen_iter);
            neighbors.erase(chosen_iter);
        }
    }
    else
        tgb.col.assign(neighbors.begin(), neighbors.end());
    tgb.row.resize(tgb.col.size(), node);
    //sampled nodes ݲҲȥأϲһ벢ȥ
    return tgb;
}

vector<NodeIDType> heads_unique(vector<NodeIDType>& array, vector<NodeIDType>& heads){
    unordered_set<NodeIDType> s(array.begin(), array.end());
#pragma omp parallel for num_threads(threads)
    for(int i=0; i<heads.size(); i++){
        if(s.count(heads[i])==1)
            s.erase(heads[i]);
    }
    array.assign(s.begin(), s.end());
    array.insert(array.begin(), heads.begin(), heads.end());
    // cout<<"s: "<<s.size()<<" array: "<<array.size()<<endl;
    return array;
}

/*------------Python Bind--------------------------------------------------------------*/
PYBIND11_MODULE(sample_cores, m)
{
    m
    .def("neighbor_sample_from_nodes", 
        &neighbor_sample_from_nodes)
    .def("get_neighbors", 
        &get_neighbors)
    .def("heads_unique", 
        &heads_unique);

    py::class_<TemporalGraphBlock>(m, "TemporalGraphBlock")
        .def(py::init<std::vector<NodeIDType> &, std::vector<NodeIDType> &,
                      std::vector<NodeIDType> &>())
        .def("row", [](const TemporalGraphBlock &tgb) { return vec2npy(tgb.row); })
        .def("col", [](const TemporalGraphBlock &tgb) { return vec2npy(tgb.col); })
        .def("nodes", [](const TemporalGraphBlock &tgb) { return vec2npy(tgb.nodes); });

    py::class_<TemporalNeighborBlock>(m, "TemporalNeighborBlock")
        .def(py::init<std::vector<vector<NodeIDType>*>&, 
                      std::vector<int> &>())
        .def_readonly("neighbors", &TemporalNeighborBlock::neighbors, py::return_value_policy::copy)
        .def_readonly("deg", &TemporalNeighborBlock::deg, py::return_value_policy::copy);
}
